/*
 * Created on Dec 22, 2005
 *
 */
package com.ibm.jps.architect.actions;

import java.text.DateFormat;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.eclipse.emf.transaction.RecordingCommand;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.gmf.runtime.notation.View;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.console.ConsolePlugin;
import org.eclipse.ui.console.IConsole;
import org.eclipse.ui.console.IConsoleManager;
import org.eclipse.ui.console.MessageConsole;
import org.eclipse.ui.console.MessageConsoleStream;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Namespace;
import org.eclipse.uml2.uml.Package;

import com.ibm.jps.architect.ArchitectPlugin;
import com.ibm.jps.architect.proc.AccessViolations;
import com.ibm.jps.architect.proc.AddPackageDependencies;
import com.ibm.jps.architect.proc.CancelException;
import com.ibm.jps.architect.proc.ObsoleteDependencies;
import com.ibm.xtools.modeler.ui.UMLModeler;

/**
 * @author jpschoch
 */
public class FixAccessViolations implements IObjectActionDelegate {
	private MessageConsoleStream consoleStream;
	private static final String CONSOLE_ICON_PATH = ""; //$NON-NLS-1$
	private static final String KEY_CONSOLE_NAME = "consoleName"; //$NON-NLS-1$
	private int totViolations;
	private int totFixedViolations;
	private int numPackagesProcessed;
	private AddPackageDependencies procAddDep;
	private ObsoleteDependencies procObsoleteDep;
	private AccessViolations procAccV;

	/**
	 * 
	 */
	public FixAccessViolations() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.IObjectActionDelegate#setActivePart(org.eclipse.jface.action.IAction,
	 *      org.eclipse.ui.IWorkbenchPart)
	 */
	public void setActivePart(IAction action, IWorkbenchPart targetPart) {

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.IActionDelegate#run(org.eclipse.jface.action.IAction)
	 */
	public void run(IAction action) {
		// Create and show the console
		showConsole();
		String undoLabel = "Fix Access Violations";
		TransactionalEditingDomain editDomain = UMLModeler.getEditingDomain();
		editDomain.getCommandStack().execute(
				new RecordingCommand(editDomain, undoLabel) {
					protected void doExecute() {
						consoleStream
								.println("+++ Starting Fix Access Violations on "
										+ getDate(new Locale("en", "US"))
										+ " +++");
						List elements = UMLModeler.getUMLUIHelper().getSelectedElements();
						totViolations = 0;
						totFixedViolations = 0;
						numPackagesProcessed = 0;
						procAccV = new AccessViolations(consoleStream);
						procAccV.setFixAccessViolations(true);
						procAddDep = new AddPackageDependencies(consoleStream);
						procObsoleteDep = new ObsoleteDependencies(consoleStream);
						for (Object curObj : elements) {
							Element modelElement = null;
							if (curObj instanceof View) {
								View view = (View) curObj;
								modelElement = (Element) view.getElement();
							} else if (curObj instanceof Element) {
								modelElement = (Element) curObj;
							}
							if (modelElement != null && modelElement instanceof Namespace) {
								Namespace ns = (Namespace) modelElement;
								try {
									// Part 1: process access violations
									consoleStream.println("\nStep 1: processing access violations");
									procAccV.execute(ns);
									totViolations = totViolations + procAccV.getTotViolations();
									totFixedViolations = totFixedViolations + procAccV.getTotFixedViolations();
									numPackagesProcessed = numPackagesProcessed + procAccV.getNumPackages();
									// Part 2: remove obsolete dependencies
									consoleStream.println("\nStep 2: removing obsolete dependendencies");
									procObsoleteDep.execute(ns);
									// Part 3: add dependencies
									if (ns instanceof Package) {
										consoleStream.println("\nStep 3: adding layer- and component-level dependencies");
										procAddDep.execute((Package) ns);
									}
								} catch (CancelException e) {
									consoleStream.println("\n!!! Fix Access Violations canceled by user");
									// MessageDialog.openInformation(shell, "Fix
									// Access Violations",
									// "Cancelled by user request.");
								}
							}
						}
						consoleStream.println("\n==> Total of "
							+ totFixedViolations + " access violation(s) resolved and "
							+ totViolations + " unresolved in "
							+ numPackagesProcessed + " package(s)");
						consoleStream.println("\n+++ End of Fix Access Violations on "
							+ getDate(new Locale("en", "US")) + " +++\n");
					}
				});
	}

	public static String getDate(Locale currentLocale) {
		Date today = new Date();
		DateFormat formatter = DateFormat.getDateTimeInstance(
				DateFormat.DEFAULT, DateFormat.DEFAULT, currentLocale);
		return formatter.format(today);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.IActionDelegate#selectionChanged(org.eclipse.jface.action.IAction,
	 *      org.eclipse.jface.viewers.ISelection)
	 */
	public void selectionChanged(IAction action, ISelection selection) {

	}

	/**
	 * Creates and display the console used to output the Show Access Violations
	 * results
	 */
	private void showConsole() {
		ImageDescriptor imageDesc = ArchitectPlugin.imageDescriptorFromPlugin(
				ArchitectPlugin.getDefault().getBundle().getSymbolicName(),
				CONSOLE_ICON_PATH);
		MessageConsole console = new MessageConsole(ArchitectPlugin
				.getResourceString(KEY_CONSOLE_NAME), imageDesc);
		IConsoleManager conMgr = ConsolePlugin.getDefault().getConsoleManager();
		conMgr.addConsoles(new IConsole[] { console });
		conMgr.showConsoleView(console);
		consoleStream = console.newMessageStream();
	}
}
