package com.ibm.jps.architect.actions;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Locale;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.transaction.RecordingCommand;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.gmf.runtime.notation.Diagram;
import org.eclipse.gmf.runtime.notation.Edge;
import org.eclipse.gmf.runtime.notation.View;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.console.ConsolePlugin;
import org.eclipse.ui.console.IConsole;
import org.eclipse.ui.console.IConsoleManager;
import org.eclipse.ui.console.MessageConsole;
import org.eclipse.ui.console.MessageConsoleStream;
import org.eclipse.uml2.uml.Association;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.ConnectableElement;
import org.eclipse.uml2.uml.Dependency;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Generalization;
import org.eclipse.uml2.uml.Interface;
import org.eclipse.uml2.uml.InterfaceRealization;
import org.eclipse.uml2.uml.Lifeline;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.UMLPackage;
import org.eclipse.uml2.uml.VisibilityKind;

import com.ibm.jps.architect.ArchitectPlugin;
import com.ibm.jps.architect.util.RSMModel;
import com.ibm.jps.architect.util.RSMRelation;
import com.ibm.jps.architect.util.RSMUtil;
import com.ibm.jps.util.ArchUtil;
import com.ibm.xtools.modeler.ui.UMLModeler;

public class ReplWithInterface implements IObjectActionDelegate {
	private MessageConsoleStream consoleStream;
	private static final String CONSOLE_ICON_PATH = ""; //$NON-NLS-1$
	private static final String KEY_CONSOLE_NAME = "consoleName"; //$NON-NLS-1$
	private boolean classModified = false;

	/**
	 * 
	 */
	public ReplWithInterface() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.IObjectActionDelegate#setActivePart(org.eclipse.jface.action.IAction,
	 *      org.eclipse.ui.IWorkbenchPart)
	 */
	public void setActivePart(IAction action, IWorkbenchPart targetPart) {

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.IActionDelegate#run(org.eclipse.jface.action.IAction)
	 */
	public void run(IAction action) {
		// Create and show the console
		showConsole();
		String undoLabel = "Promote To Interface";
		TransactionalEditingDomain editDomain = UMLModeler.getEditingDomain();
		editDomain.getCommandStack().execute(new RecordingCommand(editDomain, undoLabel) {
			protected void doExecute() {
				consoleStream.println("+++ Starting Promote To Interface on " + ArchUtil.getDate(new Locale("en", "US")) + " +++");
				final Collection selectedElements = UMLModeler.getUMLUIHelper().getSelectedElements();
				if (selectedElements.size() == 0) {
					consoleStream.println("!!! There are no selected elements");
				}
				for (Object curObj : selectedElements) {
					Class curCl = null;
					if (curObj instanceof View) {
						View view = (View) curObj;
						EObject elt = view.getElement();
						if (elt != null && elt instanceof Class) {
							curCl = (Class) elt;
						}
					} else if (curObj instanceof Class) {
						curCl = (Class) curObj;
					}
					if (curCl != null) {
						processClass(curCl);
					}
				}
				consoleStream.println("\n+++ End of Promote To Interface on " + ArchUtil.getDate(new Locale("en", "US")) + " +++\n");
			}
		});
	}


	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.IActionDelegate#selectionChanged(org.eclipse.jface.action.IAction,
	 *      org.eclipse.jface.viewers.ISelection)
	 */
	public void selectionChanged(IAction action, ISelection selection) {

	}

	/**
	 * Creates and display the console used to output the Show Access Violations
	 * results
	 */
	private void showConsole() {
		ImageDescriptor imageDesc = ArchitectPlugin.imageDescriptorFromPlugin(
				ArchitectPlugin.getDefault().getBundle().getSymbolicName(),
				CONSOLE_ICON_PATH);
		MessageConsole console = new MessageConsole(ArchitectPlugin
				.getResourceString(KEY_CONSOLE_NAME), imageDesc);
		IConsoleManager conMgr = ConsolePlugin.getDefault().getConsoleManager();
		conMgr.addConsoles(new IConsole[] { console });
		conMgr.showConsoleView(console);
		consoleStream = console.newMessageStream();
	}

	private void log(String inMsg) {
//		if (inMsg.startsWith("!!!")) {
//			RSMModel.addMarker(mRsrc, inMsg.substring(4), IMarker.SEVERITY_WARNING);
//		} else if (inMsg.startsWith("***")) {
//			RSMModel.addMarker(mRsrc, inMsg.substring(4), IMarker.SEVERITY_ERROR);
//		}
		consoleStream.println(inMsg);
	}

	@SuppressWarnings("unchecked")
	protected void processClass(Class inCl) {
		log("\n==> Processing class " + inCl.getName());
		classModified = false;
		
		// Create or retrieve the corresponding interface
		Interface lInt = getInterface(inCl);
		
		// The class must realize the interface
		List<EObject> allRea = inCl.getInterfaceRealizations();
		boolean found = false;
		for (EObject curRea_o:allRea) {
			InterfaceRealization curRea = (InterfaceRealization) curRea_o;
			Interface curInt = curRea.getContract();
			if (curInt != null && curInt == lInt) {
				found = true;
				break;
			}
		}
		if (!found) {
			InterfaceRealization lRea = inCl.createInterfaceRealization("", lInt);	
			RSMModel.setDocumentation(lRea, "*** Relation generated on " + RSMUtil.getDate(new Locale("en", "US"))
					+ " (Promote To Interface) ***");
			log("--- Created Interface Realization from '" + inCl.getName() + "' to '" + lInt.getName()+ "'");
			classModified = true;
		}

		// Remove the operations that are no longer in the class
		List<Operation> opToDel = new ArrayList<Operation>();
		for (Object curOp_o:lInt.getOperations()) {
			Operation curOp = (Operation) curOp_o;
			if (!RSMModel.classContainsOperation(inCl, curOp)) {
				String opName = curOp.getName();
				opToDel.add(curOp);
				log("--- Deleted obsolete operation '" + opName + "' from '" + lInt.getName() + "'");
				classModified = true;
			}
		}
		for (Operation curOp:opToDel) {
			curOp.destroy();
		}
		
		// The interface should have the same list of operations as the input class
		for (Object curOp_o:inCl.getOperations()) {
			Operation curOp = (Operation) curOp_o;
			if (curOp.getVisibility().equals(VisibilityKind.PUBLIC_LITERAL)) {
				if (!RSMModel.interfaceContainsOperation(lInt, curOp)) {
					RSMModel.copyOperation(curOp, lInt);
					log("--- Added public operation '" + curOp.getName() + "' to '" + lInt.getName() + "'");
					classModified = true;
				}
			}
		}

		// Look for all references to the input class and update accordingly
		processReferences(RSMModel.findClients(inCl.getModel(), inCl), inCl, lInt, true);
		processReferences(RSMModel.findSuppliers(inCl.getModel(), inCl), inCl, lInt, false);
		List<EObject> lifelines = RSMModel.findElements(inCl.getModel(), UMLPackage.eINSTANCE.getLifeline());
		for (Object obj:lifelines) {
			Lifeline lf = (Lifeline) obj;
			ConnectableElement ce = lf.getRepresents();
			if (ce != null && ce instanceof Property) {
				Property p = (Property) ce;
				if (p.getType() == inCl) {
					p.setType(lInt);
					String s = lf.getName();
					if (lf.getInteraction() != null) {
						s += " in interaction '" + lf.getInteraction().getName() + "'";
					}
					log("--- Changed type of lifeline " + s + " from '" + inCl.getName() + "' to '" + lInt.getName() + "'");			
				}
			}
		}
		// TODO views with both inCl and lInt
		
		// Wrapping up
		if (classModified) {
			log("--> Please review the changes above and either save or undo the changes.");
		} else {
			log("--> No changes.");
		}
	}

	@SuppressWarnings("unchecked")
	private Interface getInterface(Class inCl) {
		Package pkg = inCl.getNearestPackage();
		// Interface name (I<className>)
		// TODO Handle situations where the interface name is not standard (I<classname>)
		String interfaceName = "I" + inCl.getName();
		Interface lInt = null;
		
		// Check if the interface already exists:
		// - First check if the input class realizes the interface
		// - Then check if there is an interface in the same package
		// - Finally, check if there are interfaces with the same name in the whole model: if this is the case
		//   we must ask the user to select the correct interface
		List<EObject> allRea = inCl.getInterfaceRealizations();
		for (EObject curRea_o:allRea) {
			InterfaceRealization curRea = (InterfaceRealization) curRea_o;
			Interface curInt = curRea.getContract();
			if (curInt != null && curInt.getName().equals(interfaceName)) {
				lInt = curInt;
			}
		}
		
		if (lInt == null) {
			lInt = (Interface) RSMModel.findElement(inCl.getNearestPackage(), UMLPackage.eINSTANCE.getInterface(), interfaceName);
			if (lInt == null) {
				List<Element> iList = RSMModel.findElements(inCl.getModel(), UMLPackage.eINSTANCE.getInterface(), interfaceName);
				if (iList.size() == 1) {
					lInt = (Interface) iList.get(0);
				} else if (iList.size() > 1) {
					log("!!! Found several occurences of '" + interfaceName + "' (see below). This case is currently not supported: The found " +
						"\n    interfaces are ignored and a new interface will be created.");
					for (Element curElt:iList) {
						Interface curInt = (Interface) curElt;
						log("    > " + curInt.getQualifiedName());
					}
				}
			}
		}
		
		// The interface was not found: create it
		if (lInt == null) {
			lInt = (Interface) pkg.createPackagedElement("I" + inCl.getName(), UMLPackage.eINSTANCE.getInterface());
			RSMModel.setDocumentation(lInt, "*** Interface created automatically on " + RSMUtil.getDate(new Locale("en", "US"))
					+ " (Promote To Interface) ***");
			log("--- Created interface '" + interfaceName + "'");
			classModified = true;
		}
		return lInt;
	}
	
	@SuppressWarnings("unchecked")
	private void processReferences(List<RSMRelation> clients, Classifier inCl, Interface inInt, boolean isClient) {
		String label = "client";
		if (!isClient) label = "supplier";
		for (RSMRelation curRel:clients) {
			List<View> relViews = RSMModel.findViews(inCl.getModel(), curRel.getRel());
			boolean isDirty = false;
			if (curRel.getRel() instanceof Association) {
				Association assoc = (Association) curRel.getRel();
				Property prop = (Property) assoc.getMemberEnds().get(0);
				Property otherEnd = (Property) assoc.getMemberEnds().get(1);
				if (prop.getType() != inCl) {
					prop = (Property) assoc.getMemberEnds().get(1);
					otherEnd = (Property) assoc.getMemberEnds().get(0);
				}
				prop.setType(inInt);
				if (inCl.getAttributes().contains(otherEnd)) {
					try {
						inInt.getOwnedAttributes().add(otherEnd);
					} catch (Exception e) {
						e.printStackTrace();
						log("*** Error detected when attempting to move property '" + otherEnd.getName() + "' from '" + inCl.getName() + "' to '" + inInt.getName() + "': please update manually.");
					}
				}
				log("--- Changed association end for " + label + " " + RSMModel.getBasicInfo(curRel.getClient()) + " from '" + inCl.getName() + "' to '" + inInt.getName() + "'");			
				isDirty = true;
			} else if (curRel.getRel() instanceof Dependency) {
				Dependency rel = (Dependency) curRel.getRel();
				if (rel.getSuppliers().get(0) == inCl && rel.getClients().get(0) != inInt) {
					rel.getSuppliers().remove(0);
					rel.getSuppliers().add(inInt);
					NamedElement client = (NamedElement) rel.getClients().get(0); 
					client.createDependency(inCl);
					log("--- Created " + rel.eClass().getName().toLowerCase() + " from '" + client.getName() + "' to '" + inInt.getName() + "'");			
					isDirty = true;
				} else if (rel.getClients().get(0) == inCl && rel.getSuppliers().get(0) != inInt) {
					rel.getClients().remove(0);
					rel.getClients().add(inInt);
					NamedElement supplier = (NamedElement) rel.getSuppliers().get(0); 
					inCl.createDependency(supplier);
					log("--- Created " + rel.eClass().getName().toLowerCase() + " from '" + inInt.getName() + "' to '" + supplier.getName() + "'");			
					isDirty = true;
				}
			} else if (curRel.getRel() instanceof Generalization) {
				Generalization gen = (Generalization) curRel.getRel();
				if (gen.getSpecific() == inCl && gen.getGeneral() != inInt) {
					gen.setSpecific(inInt);
					log("--- Changed generalization end for " + label + " " + RSMModel.getBasicInfo(curRel.getSupplier()) + " from '" + inCl.getName() + "' to '" + inInt.getName() + "'");			
					isDirty = true;
				} else if (gen.getGeneral() == inCl && gen.getSpecific() != inInt) {
					gen.setGeneral(inInt);
					log("--- Changed generalization end for " + label + " " + RSMModel.getBasicInfo(curRel.getClient()) + " from '" + inCl.getName() + "' to '" + inInt.getName() + "'");			
					isDirty = true;
				}
			}

			// Updating views on diagrams
			if (isDirty) {
				classModified = true;
				List<Diagram> modifiedDiagrams = new ArrayList<Diagram>();
				for (View curV:relViews) {
					if (curV instanceof Edge) {
						Edge e = (Edge) curV;
						if (!modifiedDiagrams.contains(e.getDiagram())) {
							modifiedDiagrams.add(e.getDiagram());
						}
						EObject src = null;
						EObject tgt = null;
						if (e.getSource() != null) {
							src = e.getSource().getElement();
						}
						if (e.getTarget() != null) {
							tgt = e.getTarget().getElement();
						}
						if (src == inCl && tgt != inInt && tgt != inCl) {
							e.getSource().setElement(inInt);
						} else if (tgt == inCl && src != inInt && src != inCl) {
							e.getTarget().setElement(inInt);
						} 
					}
				}	
				for (Diagram curDiag:modifiedDiagrams) {
					log("    The " + curDiag.getType().toLowerCase() + " diagram '" + curDiag.getName() + "' was updated.");
				}
			}
		}
		
	}

}
