/*
 * Created on Dec 9, 2005
 *
 */
package com.ibm.jps.architect.actions;

import java.text.DateFormat;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.gmf.runtime.notation.View;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.console.ConsolePlugin;
import org.eclipse.ui.console.IConsole;
import org.eclipse.ui.console.IConsoleManager;
import org.eclipse.ui.console.MessageConsole;
import org.eclipse.ui.console.MessageConsoleStream;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.Namespace;

import com.ibm.jps.architect.ArchitectPlugin;
import com.ibm.jps.architect.proc.AccessViolations;
import com.ibm.jps.architect.proc.CancelException;
import com.ibm.xtools.modeler.ui.UMLModeler;

/**
 * @author jpschoch
 */
public class ShowAccessViolations implements IObjectActionDelegate {
	private Shell shell = null;
	private MessageConsoleStream consoleStream;
	private static final String CONSOLE_ICON_PATH = ""; //$NON-NLS-1$
	private static final String KEY_CONSOLE_NAME = "consoleName"; //$NON-NLS-1$

	/**
	 * 
	 */
	public ShowAccessViolations() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.IObjectActionDelegate#setActivePart(org.eclipse.jface.action.IAction,
	 *      org.eclipse.ui.IWorkbenchPart)
	 */
	public void setActivePart(IAction action, IWorkbenchPart targetPart) {
		shell = targetPart.getSite().getShell();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.IActionDelegate#run(org.eclipse.jface.action.IAction)
	 */
	public void run(IAction action) {
		showConsole();
		TransactionalEditingDomain editDomain = UMLModeler.getEditingDomain();
		try {
			editDomain.runExclusive(new Runnable() {
				public void run() {
					AccessViolations av = new AccessViolations(consoleStream);
					consoleStream.println("+++ Starting Show Access Violations on "
									+ getDate(new Locale("en", "US")) + " +++\n");
					List elements = UMLModeler.getUMLUIHelper().getSelectedElements();
					for (Object curObj : elements) {
						Element modelElement = null;
						if (curObj instanceof View) {
							View view = (View) curObj;
							modelElement = (Element) view.getElement();
						} else if (curObj instanceof Element) {
							modelElement = (Element) curObj;
						}
						if (modelElement != null && modelElement instanceof Namespace) {
							try {
								Namespace pkg = (Namespace) modelElement;
								consoleStream.println("==> Processing selected " + pkg.eClass().getName().toLowerCase() + " '" + pkg.getName() +"'");
								av.execute(pkg);
								int numViolations = av.getTotViolations();
								if (numViolations > 0) consoleStream.println("");
								if (pkg instanceof Model && numViolations == 0) {
									consoleStream.println("--> Total number of access violations found in '"
											+ pkg.getName() + "' = " + numViolations);
									consoleStream.println("    Consider running Fix Violations to complete processing of architectural dependencies\n");
								} else {
									consoleStream.println("--> Total number of access violations found in '"
										+ pkg.getName() + "' = " + numViolations + "\n");
								}
							} catch (CancelException e) {
								MessageDialog.openInformation(shell, "Show Access Violations", "Cancelled by user request.");
							}
						}
					}
					consoleStream.println("+++ End of Show Access Violations on "
						+ getDate(new Locale("en", "US")) + " +++\n");
				}
			});
		} catch (InterruptedException e) {
			MessageDialog.openInformation(shell, "Show Access Violations",
					"The operation was interrupted.");
		}
	}

	public static String getDate(Locale currentLocale) {
		Date today = new Date();
		DateFormat formatter = DateFormat.getDateTimeInstance(
				DateFormat.DEFAULT, DateFormat.DEFAULT, currentLocale);
		return formatter.format(today);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.IActionDelegate#selectionChanged(org.eclipse.jface.action.IAction,
	 *      org.eclipse.jface.viewers.ISelection)
	 */
	public void selectionChanged(IAction action, ISelection selection) {
	}

	/**
	 * Creates and display the console used to output the Show Access Violations
	 * results
	 */
	private void showConsole() {
		ImageDescriptor imageDesc = ArchitectPlugin.imageDescriptorFromPlugin(
				ArchitectPlugin.getDefault().getBundle().getSymbolicName(),
				CONSOLE_ICON_PATH);
		MessageConsole console = new MessageConsole(ArchitectPlugin
				.getResourceString(KEY_CONSOLE_NAME), imageDesc);
		IConsoleManager conMgr = ConsolePlugin.getDefault().getConsoleManager();
		conMgr.addConsoles(new IConsole[] { console });
		conMgr.showConsoleView(console);
		consoleStream = console.newMessageStream();
	}
}
