package com.ibm.jps.architect.actions;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Locale;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.gmf.runtime.notation.Diagram;
import org.eclipse.gmf.runtime.notation.Edge;
import org.eclipse.gmf.runtime.notation.View;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.console.MessageConsoleStream;
import org.eclipse.uml2.uml.Actor;
import org.eclipse.uml2.uml.Association;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.ConnectableElement;
import org.eclipse.uml2.uml.Dependency;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Generalization;
import org.eclipse.uml2.uml.Lifeline;
import org.eclipse.uml2.uml.Message;
import org.eclipse.uml2.uml.MessageSort;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.Namespace;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.Type;

import com.ibm.jps.architect.util.RSMRelation;
import com.ibm.jps.util.ArchUtil;
import com.ibm.xtools.modeler.ui.UMLModeler;
import com.ibm.xtools.uml.ui.diagram.IUMLDiagramHelper;

public class ShowMissingRelations implements IObjectActionDelegate {
	private Shell shell = null;
	private MessageConsoleStream consoleStream = null;
	private boolean headerPrinted;
	private Integer numMissingRelations;

	public ShowMissingRelations() {
		super();
	}

	public void setActivePart(IAction action, IWorkbenchPart targetPart) {
		shell = targetPart.getSite().getShell();
	}

	public void run(IAction action) {
		// Create and show the console
		consoleStream = ArchUtil.showConsole();
		TransactionalEditingDomain editDomain = UMLModeler.getEditingDomain();
		try {
			editDomain.runExclusive(new Runnable() {
				public void run() {
					consoleStream.println("+++ Starting Show Missing Relations on " + ArchUtil.getDate(new Locale("en", "US")) + " +++");

					final Collection selectedElements = UMLModeler.getUMLUIHelper().getSelectedElements();
					if (selectedElements.size() == 0) {
						consoleStream.println("!!! There are no selected elements");
					}
					for (Object curObj : selectedElements) {
						boolean processed = true;
						numMissingRelations = 0;
						if (curObj instanceof Diagram) {
							Diagram d = (Diagram) curObj;
							log("\n==> Processing diagram " + d.getName());
							processDiagram(d);
						} else if (curObj instanceof View) {
							View view = (View) curObj;
							EObject elt = view.getElement();
							processed = false;
							if (elt != null) {
								if (elt instanceof Diagram) {
									Diagram d = (Diagram) elt;
									log("\n==> Processing diagram " + d.getName());
									processDiagram(d);
									processed = true;
								} else if (elt instanceof Namespace) {
									log("\n==> Processing " + getElementBasicInfo(elt));
									processNamespace((Namespace) elt);
									processed = true;
								}
							}
						} else if (curObj instanceof Namespace) {
							log("\n==> Processing " + getElementBasicInfo((EObject) curObj));
							processNamespace((Namespace) curObj);
						}
						if (processed) {
							if (numMissingRelations > 0) {
								log("\n--> " + numMissingRelations + " missing relationship(s) found");									
							} else {
								log("--> " + "No missing relationship(s) found");	
							}
						}
					}
					consoleStream.println("\n+++ End of Show Missing Relations on " + ArchUtil.getDate(new Locale("en", "US"))+ " +++\n");
				}
			});
		} catch (InterruptedException e) {
			MessageDialog.openInformation(shell, "Show Missing Relations", "The operation was interrupted.");
		}
	}

	public void selectionChanged(IAction action, ISelection selection) {
		// TODO Auto-generated method stub

	}

	protected void processDiagram(Diagram inDiag) {
		headerPrinted = false;
		for (Object curEdge_o:inDiag.getEdges()) {
			Edge curEdge = (Edge) curEdge_o;
			EObject elt = curEdge.getElement();
			if (elt != null && elt instanceof Message) {
				processMessage((Message) elt, curEdge, inDiag);
			}
		}
	}

	private void processMessage(Message inMsg, Edge inEdge, Diagram inDiag) {
		int msgSort = inMsg.getMessageSort().getValue();
		if (msgSort == MessageSort.SYNCH_CALL ||
			msgSort == MessageSort.ASYNCH_CALL ||
			msgSort == MessageSort.CREATE_MESSAGE || 
			msgSort == MessageSort.DELETE_MESSAGE) {
			Classifier src = getSourceClassifier(inEdge);
			Classifier tgt = getTargetClassifier(inEdge);
			if (src != null && tgt != null && !(src instanceof Actor) && !(tgt instanceof Actor) && src != tgt) {
				boolean relationExists = false;
				List<RSMRelation> rels = getRelations(src);
				for (RSMRelation curRel:rels) {
					if (tgt.equals(curRel.getSupplier())) {
						relationExists = true;
						break;
					}
				}
				if (!relationExists) {
					if (!headerPrinted) {
						log("\n--- Processing Diagram " + inDiag.getType() + " " + inDiag.getName());
						headerPrinted = true;
					}
					numMissingRelations++;
					log("*** Missing Relationship > " + inMsg.getMessageSort().getLiteral() + " message '" + inMsg.getName() + "' from " + getElementBasicInfo(src) + " to " + getElementBasicInfo(tgt));
				}
			}
		}
	}

	private Classifier getSourceClassifier(Edge inEdge) {
		return extractClassifier(inEdge.getSource());
	}
	
	private Classifier getTargetClassifier(Edge inEdge) {
		return extractClassifier(inEdge.getTarget());
	}
	
	private Classifier extractClassifier(View inView) {
		if (inView != null) {
			EObject e = inView.getElement();
			if (e != null && e instanceof Lifeline) {
				Lifeline l = (Lifeline) e;
				ConnectableElement c = l.getRepresents();
				if (c != null && c instanceof Property) {
					Property p = (Property) c;
					Type t = p.getType();
					if (t != null && t instanceof Classifier) {
						return (Classifier) t;
					}
				}
			}
		}
		return null;		
	}

    public static List<RSMRelation> getRelations(Classifier cl) {
        List<RSMRelation> allRels = new ArrayList<RSMRelation>();
        // Generalizations
        for (Object curGen_o:cl.getGeneralizations()) {
            Generalization curGen = (Generalization) curGen_o;
            RSMRelation rel = new RSMRelation();
            rel.setClient(cl);
            rel.setRel(curGen);
            for (Object curObj:curGen.getTargets()) {
                rel.setSupplier((NamedElement) curObj);
                break;
            }
            allRels.add(rel);
        }
        // Dependencies
        for (Object curDep_o:cl.getClientDependencies()) {
            Dependency curDep = (Dependency) curDep_o;
            RSMRelation rel = new RSMRelation();
            rel.setClient(cl);
            rel.setRel(curDep);
            for (Object curObj:curDep.getSuppliers()) {
                rel.setSupplier((NamedElement) curObj);
                break;
            }
            allRels.add(rel);
        }
        // Attributes and relations
        for (Object curElt_o:cl.getOwnedElements()) {
            Element curElt = (Element) curElt_o;
            if (curElt instanceof Property) {
                Property prop = (Property) curElt;
                Association assoc = prop.getAssociation();
                if (assoc != null) {
                    RSMRelation rel = new RSMRelation();
                    rel.setClient(cl);
                    rel.setRel(assoc);
                    for (Object curType_o:assoc.getEndTypes()) {
                        Type curType = (Type) curType_o;
                        if (!cl.getQualifiedName().equals(curType.getQualifiedName())) {
                            rel.setSupplier(curType);
                            break;
                        }
                    }
                    allRels.add(rel);
                }
            }
        }
        return allRels;
    }

	protected void processNamespace(Namespace inNS) {
		// log(getElementBasicInfo(inNS));
		IUMLDiagramHelper helper = UMLModeler.getUMLDiagramHelper();
		for (Object o : helper.getDiagrams((Namespace) inNS)) {
			processDiagram((Diagram) o);
		}
		for (Object o : inNS.getOwnedElements()) {
			if (o instanceof Namespace) {
				processNamespace((Namespace) o);
			}
		}
	}

	private void log(String inMsg) {
//		if (inMsg.startsWith("!!!")) {
//			RSMModel.addMarker(mRsrc, inMsg.substring(4), IMarker.SEVERITY_WARNING);
//		} else if (inMsg.startsWith("***")) {
//			RSMModel.addMarker(mRsrc, inMsg.substring(4), IMarker.SEVERITY_ERROR);
//		}
		consoleStream.println(inMsg);
	}

	private String getElementBasicInfo(EObject elt) {
		String s = "";
		if (elt != null) {
			s = elt.getClass().getSimpleName();
			if (s.endsWith("Impl"))
				s = s.substring(0, s.length() - "Impl".length());
			else
				s = elt.getClass().getSimpleName();
			if (elt instanceof NamedElement) {
				NamedElement ne = (NamedElement) elt;
				if (ne.getName() == null || ne.getName().equals(""))
					s += " <unnamed>";
				else
					s += " '" + ne.getName() + "'";
				if (ne instanceof Message) {
					Message msg = (Message) ne;
					s += " (" + msg.getMessageSort().getLiteral() + ")";
				}
			}
		} else {
			s = "element = <null>";
		}
		return s;
	}

}
