/*
 * Created on Dec 22, 2005
 *
 */
package com.ibm.jps.architect.proc;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.ui.console.MessageConsoleStream;
import org.eclipse.uml2.uml.Dependency;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.PackageImport;
import org.eclipse.uml2.uml.Profile;
import org.eclipse.uml2.uml.ProfileApplication;
import org.eclipse.uml2.uml.Stereotype;

import com.ibm.jps.architect.util.RSMModel;
import com.ibm.jps.architect.util.RSMRelation;
import com.ibm.jps.architect.util.RSMUtil;

/**
 * @author jpschoch
 */
public class AddPackageDependencies {
    private MessageConsoleStream console = null;
    private boolean isDebug = true;
    private Model targetModel = null;
    private int rootLevel;
    private int layerlevel;
    private int componentLevel;
    private int numDepAdded;
    private int numPkgProcessed;
    private List<String> newDependencies = new ArrayList<String>();
    private List<String> externalLibraries = new ArrayList<String>();
    private List<String> importedLibraries = new ArrayList<String>();
    private List<Package> neededDepSrc = new ArrayList<Package>();
    private List<Package> neededDepTgt = new ArrayList<Package>();
    private List<String> neededDepDoc = new ArrayList<String>();
    private int curLevel = 0;
    private int refLevel = 0;
    private String curPackageName = "";
    private Package curPackage = null;
    private IResource mRsrc = null;

    public AddPackageDependencies(MessageConsoleStream inConsole) {
        super();
        this.console = inConsole;
    }

    private void log(String inStr) {
        if (inStr.startsWith("!!!")) {
            RSMModel.addMarker(mRsrc, inStr.substring(4), IMarker.SEVERITY_WARNING);
        } else if (inStr.startsWith("***")) {
            RSMModel.addMarker(mRsrc, inStr.substring(4), IMarker.SEVERITY_ERROR);
        }
        console.println(inStr);
        if (isDebug) System.out.println(inStr);
    }

    private void determineLevels() {
        layerlevel = -1;
        checkForLayerLevel(targetModel, 0);
        if (layerlevel == -1) {
            log("!!! No <<layer>> packages detected: using top level as layer level");
            layerlevel = 1;
        }
        rootLevel = layerlevel - 1;
        componentLevel = layerlevel + 1;
    }

    /**
     * @param package1
     */
    @SuppressWarnings("unchecked")
	private void checkForLayerLevel(Package inPkg, int inCurL) {
        Stereotype st = inPkg.getAppliedStereotype("ArchitectAssistant::layer");
        if (st != null) {
            if (layerlevel == -1)
                layerlevel = inCurL;
            else if (inCurL != layerlevel)
                    log("!!! '" + inPkg.getQualifiedName() + "' at a lower level than other(s) <<layer>> package(s).");
        }
        st = inPkg.getAppliedStereotype("ArchitectAssistant::externalLibrary");
        if (st != null) {
            externalLibraries.add(RSMModel.getNameWithoutModel(inPkg.getQualifiedName()));
        }
        st = inPkg.getAppliedStereotype("Standard::modelLibrary");
        if (st != null) {
            externalLibraries.add(RSMModel.getNameWithoutModel(inPkg.getQualifiedName()));
        }
        List<Object> imports = inPkg.getPackageImports();
        for (Object curPI_o : imports) {
        	PackageImport curPI = (PackageImport) curPI_o;
        	Package importedPkg = curPI.getImportedPackage();
            if (importedPkg.isModelLibrary()) {
                importedLibraries.add(RSMModel.getNameWithoutModel(importedPkg.getQualifiedName()));
            }
        }
        for (Object curPkg_o : inPkg.getNestedPackages()) {
            checkForLayerLevel((Package) curPkg_o, inCurL + 1);
        }
    }

    public void execute(Package inPkg) throws CancelException {
        mRsrc = RSMModel.getModelResource(inPkg);
        // The target model should have the ArchitectAssistant profile applied to it.
        boolean profileApplied = false;
        if (inPkg instanceof Model)
            targetModel = (Model) inPkg;
        else
            targetModel = inPkg.getModel();
        for (Object curPA_o : targetModel.getProfileApplications()) {
			ProfileApplication curPA = (ProfileApplication) curPA_o;
			Profile profile = curPA.getAppliedProfile();
            if (profile.getQualifiedName().compareTo("ArchitectAssistant") == 0) {
                profileApplied = true;
                break;
            }
        }
        if (!profileApplied) {
            log("--- Profile 'ArchitectAssistant' not applied");
            RSMModel.addMarker(mRsrc, "Profile ArchitectAssistant not applied.", IMarker.SEVERITY_INFO);
        } else {
            determineLevels();
            if (isDebug) System.out.println(">>> AddPackageDependencies.execute: Component level");
            refLevel = componentLevel;
            neededDepSrc.clear();
            neededDepTgt.clear();
            neededDepDoc.clear();
            processPackage(inPkg);
            for (Package curSrc : neededDepSrc) {
            	int index = neededDepSrc.indexOf(curSrc);
    	        Dependency dep = curSrc.createDependency(neededDepTgt.get(index));
    	        RSMModel.setDocumentation(dep, neededDepDoc.get(index));
            }
            if (isDebug) System.out.println(">>> AddPackageDependencies.execute: Layer level");
            refLevel = layerlevel;
            neededDepSrc.clear();
            neededDepTgt.clear();
            neededDepDoc.clear();
            processPackage(inPkg);
            for (Package curSrc : neededDepSrc) {
            	int index = neededDepSrc.indexOf(curSrc);
    	        Dependency dep = curSrc.createDependency(neededDepTgt.get(index));
    	        RSMModel.setDocumentation(dep, neededDepDoc.get(index));
            }
        }
    }

    /**
     * @param inPkg
     * @throws CancelException
     */
    private void processPackage(Package inPkg) throws CancelException {
        if (isDebug) System.out.println("\n==> Processing package " + inPkg.getQualifiedName());
        numPkgProcessed++;
        // Only check package relationships if the package is at the specified refLevel
        // (component or layer)
        int cur = getLevel(inPkg, true);
        if (cur >= refLevel) checkPackageRelationships(inPkg);
        // Process nested packages
        for (Object curPkg_o : inPkg.getNestedPackages()) {
            processPackage((Package) curPkg_o);
        }
    }

    private boolean isExternal(Package inPkg) {
        if ((inPkg.getAppliedStereotype("ArchitectAssistant::externalLibrary") != null) ||
        		(inPkg.getAppliedStereotype("Standard::modelLibrary") != null)) 
        	return true;
        return false;
    }

    private int getLevel(Package inPkg, boolean inSetCurVariables) {
        String pkgName = RSMModel.getNameWithoutModel(inPkg.getQualifiedName());
        int pos = 0;
        int prev = -1;
        int level;
        if (isExternal(inPkg))
            level = -1;
        else {
            level = 0;
            while (pos >= 0) {
                pos = pkgName.indexOf("::", prev + 2);
                prev = pos;
                level++;
            }
            if (level < rootLevel) level = -1;
        }
        if (inSetCurVariables && level == refLevel) {
            curLevel = level;
            curPackage = inPkg;
            if (prev > 2)
                curPackageName = pkgName.substring(0, prev);
            else if (prev == -1)
                curPackageName = pkgName;
            else
                curPackageName = "";
        }
        return level;
    }

    private String extractParentPackageName(Package inFromPkg, int inAtLevel) {
        for (int i = 0; i < importedLibraries.size(); i++) {
            String s = importedLibraries.get(i).toString();
            if (inFromPkg.getQualifiedName().equals(s) || inFromPkg.getQualifiedName().startsWith(s + "::")) return s;
        }
        String pkgName = RSMModel.getNameWithoutModel(inFromPkg.getQualifiedName());

        // First check for external libraries
        for (int i = 0; i < externalLibraries.size(); i++) {
            String s = externalLibraries.get(i).toString();
            if (pkgName.equals(s) || pkgName.startsWith(s + "::")) return s;
        }
 
        // Move to the requested level
        int pos = 0;
        int prev = -1;
        int level = 0;
        int refLev = inAtLevel;
        while (pos >= 0) {
            pos = pkgName.indexOf("::", prev + 2);
            if (pos >= 0) prev = pos;
            level++;
            if (level == refLev) {
                if (pos > 0)
                    return pkgName.substring(0, pos);
                else
                    return pkgName;
            }
        }

        // The search was unsuccessful: return the full package name
        return pkgName;
    }

    private Package findParentPackage(Package inForPkg, String inWithName) {
        if (RSMModel.getNameWithoutModel(inForPkg.getQualifiedName()).equals(inWithName)) return inForPkg;
        Package cur = inForPkg;
        Package container;
        while (true) {
            container = cur.getNestingPackage();
            if (container != null) {
                if (RSMModel.getNameWithoutModel(container.getQualifiedName()).equals(inWithName))
                    return container;
                else
                    cur = container;
            } else
                return null;
        }
    }

    private Package curLevelPackage(Package inPkg) {
        String pkgName = RSMModel.getNameWithoutModel(inPkg.getQualifiedName());
        if (pkgName.startsWith(curPackageName)) {
            if (isDebug) System.out.println("    1 - " + curPackageName + " <-- " + inPkg.getQualifiedName());
            return curPackage;
        } else {
            String parentName = extractParentPackageName(inPkg, curLevel);
            if (isDebug) System.out.println("    2 - " + parentName + " <-- " + inPkg.getQualifiedName());
            // Starting from the input package, go back to the container, then to the container's
            // container until we find the parent whose name matches parentName
            return findParentPackage(inPkg, parentName);
        }
    }

    private boolean dependencyExists(Package inFrom, Package inTo) {
        if (!newDependencies.isEmpty()) {
            String depName = RSMModel.getNameWithoutModel(inFrom.getQualifiedName()) + ">"
                    + RSMModel.getNameWithoutModel(inTo.getQualifiedName());
            if (newDependencies.contains(depName)) return true;
        }
        return RSMModel.dependencyExists(inFrom, inTo);
    }

    private void checkPackageRelationships(Package inClient) throws CancelException {
        try {
            List<RSMRelation> allRel = RSMModel.getDependencies(inClient);
            for (RSMRelation rel : allRel) {
                NamedElement supplier = rel.getSupplier();
                String reason = rel.getRel().eClass().getName().toLowerCase() + " from " + inClient.getQualifiedName()
                        + " to " + supplier.getQualifiedName();
                if (isDebug) System.out.println("A. " + reason);
                // Only look at relations between classes and relations between packages
                Package clientPkg = inClient.getNearestPackage();
                Package supplierPkg = supplier.getNearestPackage();
                if (clientPkg != null && supplierPkg != null) {
                    clientPkg = curLevelPackage(clientPkg);
                    supplierPkg = curLevelPackage(supplierPkg);
                }
                // At this stage , there is a relationship between a class from the package and
                // another class, or
                // between a package inside the package and another package. Because top packages
                // are
                // contained within models, make sure the containers are packages.
                if (clientPkg != null && supplierPkg != null) {
                    // The client must be the current package or one of its sub-packages
                    if (RSMModel.getNameWithoutModel(inClient.getQualifiedName()).startsWith(curPackageName)) {
                        // At least one of the packages must be "below" the level of the current
                        // package.
                        int clientLevel = getLevel(clientPkg, false);
                        int supplierLevel = getLevel(supplierPkg, false);
                        if (isDebug) System.out.println("B. " + rel.getRel().eClass().getName().toLowerCase() + " from "
                                        + inClient.eClass().getName().toLowerCase() + " in "
                                        + RSMModel.getNameWithoutModel(clientPkg.getQualifiedName()) + " (lev. "
                                        + clientLevel + ") to " + supplier.eClass().getName().toLowerCase() + " in "
                                        + RSMModel.getNameWithoutModel(supplierPkg.getQualifiedName()) + " (lev. "
                                        + supplierLevel + ")");
                        if (clientLevel >= curLevel || supplierLevel >= curLevel || supplierLevel == -1) {
                            // For packages below the level of the current package, identify the
                            // parent package with the same level
                            // as the current package
                            if (isDebug) System.out.println("C. candidate "
                                            + RSMModel.getNameWithoutModel(clientPkg.getQualifiedName()) + " to "
                                            + RSMModel.getNameWithoutModel(supplierPkg.getQualifiedName()));
                            if (!dependencyExists(clientPkg, supplierPkg))
                                    createDependency(clientPkg, supplierPkg, reason);
                        }
                    }
                }
            }
        } catch (Exception e) {
            log("*** AddPackageDependencies.checkPackageRelationships: " + e.getMessage());
        }
    }

    private void createDependency(Package inFrom, Package inTo, String inReason) {
        String fromName = RSMModel.getNameWithoutModel(inFrom.getQualifiedName());
        String toName = RSMModel.getNameWithoutModel(inTo.getQualifiedName());
        String depName = fromName + ">" + toName;
        if (!newDependencies.contains(depName)) {
	        log("+++ Adding dependency from '" + inFrom.getQualifiedName() + "' to '" + inTo.getQualifiedName() + "'");
	        // Processing deferred
	        neededDepSrc.add(inFrom);
	        neededDepTgt.add(inTo);
	        neededDepDoc.add("*** Relation generated on " + RSMUtil.getDate(new Locale("en", "US"))
	                + ", based on " + inReason + " (AddPackageDependencies) ***");
	        newDependencies.add(fromName + ">" + toName);
	        numDepAdded++;
        }
    }

    /**
     * @return
     */
    public int getNumDepAdded() {
        return numDepAdded;
    }

    /**
     * @return
     */
    public int getNumPkgProcessed() {
        return numPkgProcessed;
    }
}
