/*
 * Created on Dec 22, 2005
 *
 */
package com.ibm.jps.architect.proc;

import java.util.List;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.console.MessageConsoleStream;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.Namespace;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.Relationship;
import org.eclipse.uml2.uml.UMLPackage;

import com.ibm.jps.architect.util.RSMModel;
import com.ibm.jps.architect.util.RSMRelation;


/**
 * @author jpschoch
 */
public class ObsoleteDependencies {
    private MessageConsoleStream console = null;
    private Shell shell = null;
    private boolean isDebug = true;
    private IResource mRsrc = null;

    public ObsoleteDependencies(MessageConsoleStream inConsole) {
        super();
        this.shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
        this.console = inConsole;
    }

    private void log(String inStr) {
        if (inStr.startsWith("!!!")) {
            RSMModel.addMarker(mRsrc, inStr.substring(4), IMarker.SEVERITY_WARNING);
        } else if (inStr.startsWith("***")) {
            RSMModel.addMarker(mRsrc, inStr.substring(4), IMarker.SEVERITY_ERROR);
        }
        console.println(inStr);
        if (isDebug) System.out.println(inStr);
    }

    public void execute(Namespace inNS) throws CancelException {
        processNS(inNS);
    }

    /**
     * @param inNS
     * @throws CancelException
     */
    private void processNS(Namespace inNS) throws CancelException {
        if (isDebug) System.out.println("\n>>> Processing " + RSMModel.getBasicInfo(inNS));
        processNamespaceDependencies(inNS);
        // Process nested packages
        if (inNS instanceof Package) {
	        for (Object curPkg_o : ((Package) inNS).getNestedPackages()) {
	            processNS((Package) curPkg_o);
	        }
        }
    }

    /**
     * processPackageDependencies checks the dependencies of the input package with all other
     * packages. For each dependency, it checks whether or not the dependency is still justified. If
     * this is not the case, the dependency is automatically removed if it was script generated,
     * otherwise the script prompts the user to determine the correct action to take.
     * 
     * @param IRXEPackage
     *            pkg
     */
    private void processNamespaceDependencies(Namespace inNS) throws CancelException {
        List<RSMRelation> deps = RSMModel.getDependencies(inNS);
        for (RSMRelation curRel : deps) {
            NamedElement supplier = curRel.getSupplier();
            // Only look at dependencies to other packages (not dependencies from other packages
            // to this one).
            if (supplier != null && supplier.eClass() == UMLPackage.eINSTANCE.getPackage()) {
                if (isDebug) System.out.println(">>> processNamespaceDependencies: dependency from " + inNS.getName() + " to "
                   + supplier.getName() + " (" + supplier.getQualifiedName() + ")");
                Package targetPkg = (Package) supplier;
                if (inNS.getAppliedStereotype("ArchitectAssistant::layer") == null || targetPkg.getAppliedStereotype("ArchitectAssistant::layer") == null) {
                    if (!RSMModel.areNamespacesDependent(inNS, targetPkg)) {
                        Relationship dep = curRel.getRel();
                        String relLabel = dep.eClass().getName().toLowerCase() + " from '" + inNS.getQualifiedName()
                                + "' to '" + targetPkg.getQualifiedName() + "'";
                        String s = RSMModel.getDocumentation(dep);
                        String reason = "";
                        boolean scriptGenerated = (s.startsWith("*** Relation generated on ") && ((s
                                .indexOf("AccessViolations") > 0) || (s.indexOf("AddPackageDependencies") > 0)));
                        boolean deleteIt = false;
                        if (scriptGenerated) {
                            deleteIt = true;
                            reason = "--- Script-generated dependency no longer required, auto-deleted";
                        } else {
                            // Si c'est une dpendance cre manuellement => demander avant
                            // de la supprimer
                            MessageBox box = new MessageBox(shell, SWT.YES | SWT.NO | SWT.CANCEL
                                    | SWT.APPLICATION_MODAL | SWT.ICON_QUESTION);
                            box.setText("ObsoleteDependencies Module");
                            box.setMessage("There is a dependency that is no longer required from "
                                    + inNS.eClass().getName().toLowerCase() + ":\n\n   '"
                                    + RSMModel.getNameWithoutModel(inNS.getQualifiedName()) + "'\n\nto "
                                    + targetPkg.eClass().getName().toLowerCase() + ":\n\n   '"
                                    + RSMModel.getNameWithoutModel(targetPkg.getQualifiedName())
                                    + "'.\n\nDo you want to delete it?");
                            int reply = box.open();
                            if (reply == SWT.YES) {
                                deleteIt = true;
                                reason = "--- Dependency not required, deleted by user request";
                            } else if (reply == SWT.CANCEL)
                                throw new CancelException();
                            else {
                                reason = "--- Dependency not required, maintained by user request";
                            }
                        }
                        if (deleteIt) {
                            dep.destroy();
                        } else {
                            if (scriptGenerated)
                                reason = "--- Script-generated dependency no longer required";
                            else
                                reason = "--- Dependency not required";
                        }
                        log(reason + ": " + relLabel);
                    }
                }
            }
        }
    }
}
