/*
 * Created on Nov 18, 2005
 *
 */
package com.ibm.jps.architect.util;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.gmf.runtime.notation.Edge;
import org.eclipse.gmf.runtime.notation.Node;
import org.eclipse.gmf.runtime.notation.NotationPackage;
import org.eclipse.gmf.runtime.notation.View;
import org.eclipse.uml2.uml.Association;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Comment;
import org.eclipse.uml2.uml.Dependency;
import org.eclipse.uml2.uml.DirectedRelationship;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Generalization;
import org.eclipse.uml2.uml.Interface;
import org.eclipse.uml2.uml.Message;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.Namespace;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.Parameter;
import org.eclipse.uml2.uml.ParameterDirectionKind;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.Stereotype;
import org.eclipse.uml2.uml.Type;
import org.eclipse.uml2.uml.UMLPackage;

import com.ibm.xtools.modeler.ui.UMLModeler;


/**
 * @author jpschoch
 */
public class RSMModel {

    public static String getDocumentation(Element from) {
        for (Iterator i = from.getOwnedComments().iterator(); i.hasNext();) {
            Comment c = (Comment) i.next();
            if (c.getAppliedStereotype("Default::Documentation") != null) {
				return c.getBody();
            }
        }
        return "";
    }


    public static NamedElement getElementByQualifiedName(Package inPkg, String elementQualifiedName) {
        NamedElement elt = getElementByQualifiedNameFromPkg(inPkg, elementQualifiedName);
        if (elt != null) return elt;
        for (Iterator i = inPkg.getNestedPackages().iterator(); i.hasNext();) {
            elt = getElementByQualifiedName((Package) i.next(), elementQualifiedName);
            if (elt != null) return elt;
        }
        return null;
    }

    private static NamedElement getElementByQualifiedNameFromPkg(Package inPkg, String elementQualifiedName) {
        for (Iterator i = inPkg.getOwnedElements().iterator(); i.hasNext();) {
            Object obj = i.next();
            if (obj instanceof NamedElement) {
                NamedElement element = (NamedElement) obj;
                if (element.getQualifiedName().equals(elementQualifiedName)) {
                    return element;
                }
            }
        }
        return null;
    }

    public static void setDocumentation(NamedElement to, String text) {
        Comment tc = null;
        for (Iterator i = to.getOwnedComments().iterator(); i.hasNext();) {
            Comment c = (Comment) i.next();
            if (c.getAppliedStereotype("Default::Documentation") != null) {
                tc = c;
                break;
            }
        }
        if (tc == null) {
            tc = to.createOwnedComment();
            tc.applyStereotype(tc.getApplicableStereotype("Default::Documentation"));
        }
		tc.setBody(text);
    }

    public static String getStereotypes(Element elt) {
        String res = "";
        for (Iterator i = elt.getAppliedStereotypes().iterator(); i.hasNext();) {
            Stereotype st = (Stereotype) i.next();
            if (res.length() == 0)
                res = "<<";
            else
                res = res + ", ";
            res = res + st.getName();
        }
        if (res.length() > 0) res = res + ">>";
        return res;
    }

    public static String getNameWithoutModel(String name) {
        int pos = name.indexOf("::");
        if (pos > 0)
            return name.substring(name.indexOf("::") + 2);
        else
            return name;
    }

    /**
     * containedIn determines if the 'source' package is contained in the 'target' package.
     * 
     * @param String
     *            source
     * @param String
     *            target
     * @return boolean
     */
    public static boolean containedIn(String source, String target) {
        return (source.startsWith(target + "::"));
    }

    /**
     * dependencyExists returns true if there is a dependency relationship whether explicit or
     * implicit.
     */
    public static boolean dependencyExists(Namespace inFrom, Namespace inTo) {
    	if (isGlobal(inTo)) return true;
        String fromName = getNameWithoutModel(inFrom.getQualifiedName());
        String toName = getNameWithoutModel(inTo.getQualifiedName());
        if (fromName.equals(toName)) return true;
        // Supplier contained in client
        if (containedIn(toName, fromName)) return true;
        // Check if the client has a dependency to the supplier
        for (Object curDep_o : inFrom.getClientDependencies()) {
            Dependency curDep = (Dependency) curDep_o;
            for (Object curSup_o : curDep.getSuppliers()) {
                if (curSup_o == inTo) return true;
            }
        }
        return false;
    }

    public static List<RSMRelation> getDependencies(NamedElement inNE) {
        List<RSMRelation> allRels = new ArrayList<RSMRelation>();
        // Dependencies
        for (Object curDep_o : inNE.getClientDependencies()) {
            Dependency curDep = (Dependency) curDep_o;
            RSMRelation rel = new RSMRelation();
            rel.setClient(inNE);
            rel.setRel(curDep);
            for (Object curSup_o : curDep.getSuppliers()) {
                rel.setSupplier((NamedElement) curSup_o);
                break;
            }
            allRels.add(rel);
        }
        return allRels;
    }

    public static List<RSMRelation> getRelations(Classifier inCl) {
        List<RSMRelation> allRels = new ArrayList<RSMRelation>();
        // Generalizations
        for (Object curGen_o : inCl.getGeneralizations()) {
            Generalization curGen = (Generalization) curGen_o;
            RSMRelation rel = new RSMRelation();
            rel.setClient(inCl);
            rel.setRel(curGen);
            for (Object curTgt_o : curGen.getTargets()) {
                rel.setSupplier((NamedElement) curTgt_o);
                break;
            }
            allRels.add(rel);
        }
        // Dependencies
        for (Object curDep_o : inCl.getClientDependencies()) {
            Dependency curDep = (Dependency) curDep_o;
            RSMRelation rel = new RSMRelation();
            rel.setClient(inCl);
            rel.setRel(curDep);
            for (Object curSupp_o : curDep.getSuppliers()) {
                rel.setSupplier((NamedElement) curSupp_o);
                break;
            }
            allRels.add(rel);
        }
        // Attributes and relations
        for (Object curElt_o : inCl.getOwnedElements()) {
            Element curElt = (Element) curElt_o;
            if (curElt instanceof Property) {
                Property prop = (Property) curElt;
                Association assoc = prop.getAssociation();
                if (assoc != null) {
                    RSMRelation rel = new RSMRelation();
                    rel.setClient(inCl);
                    rel.setRel(assoc);
                    String s = getStereotypes(assoc);
                    if (s.length() > 0) s = s + " ";
                    for (Object curType_o : assoc.getEndTypes()) {
                        Type curType = (Type) curType_o;
                        if (inCl != curType) {
                            rel.setSupplier(curType);
                            break;
                        }
                    }
                    allRels.add(rel);
                }
            }
        }
        return allRels;
    }

    /**
     * arePackagesDependent determines whether or not the package 'from' depends on the package
     * 'to'.
     * 
     * @return boolean
     */
    public static boolean areNamespacesDependent(Namespace inFrom, Namespace inTo) {
        String fromName = getNameWithoutModel(inFrom.getQualifiedName());
        String toName = getNameWithoutModel(inTo.getQualifiedName());
        
        // If the supplier package (inTo) and the client package (inFrom) are the same object
        // or the supplier (inTo) is contained in the client (inFrom), the dependency is
        // implied. Return 'false' so that the explicit dependency can be suppressed.
        if (fromName.equals(toName) || containedIn(toName, fromName)) return false;

        // First check dependencies based on elements directly container within the source
        // package. If this does not work, check all nested packages within the source
        // package.
        if (namespaceDependsOn(inFrom, inTo)) return true;

        if (inFrom instanceof Package) {
	        for (Object curPkg_o : ((Package) inFrom).getNestedPackages()) {
	            if (areNamespacesDependent((Package) curPkg_o, inTo)) return true;
	        }
        }
        return false;
    }

    /**
     * dependsOn determines if the source namespace (inFrom) depends on the target namespace (inTo),
     * i.e. if the source namespace contains any element in relationship with an element in the target
     * namespace.
     */
    private static boolean namespaceDependsOn(Namespace inFrom, Namespace inTo) {
        for (Object curElt_o : inFrom.getOwnedElements()) {
            Element curElt = (Element) curElt_o;
            if (curElt instanceof Classifier) {
                if (classifierDependsOn((Classifier) curElt, inTo)) return true;
            }
        }
        return false;
    }

    /**
     * dependsOn determines if the input element (inElt) depends on the target package (inPkg), 
     * i.e. if inElt has a relationship with any element belonging to the target package. Note that only
     * classes and components are considered at this stage. Packages are not included because they
     * are expected to have dependencies on other packages only as a result of class or component
     * dependencies.
     */
    public static boolean classifierDependsOn(Classifier inElt, Namespace inNS) {
        // Process relations
        List<RSMRelation> rels = getRelations(inElt);
        for (RSMRelation rel : rels) {
            NamedElement supplier = rel.getSupplier();
            if (supplier != null && belongsTo(supplier, inNS)) return true;
        }
        // Process attributes
        for (Object curProp_o : inElt.getAttributes()) {
            Property curProp = (Property) curProp_o;
            if (curProp.getAssociation() == null) {
                Type attType = curProp.getType();
                if (attType != null && belongsTo(attType, inNS)) return true;
            }
        }
        // Process operations
        if (inElt instanceof Class || inElt instanceof Interface) {
            List operations;
            if (inElt instanceof Class)
                operations = ((Class) inElt).getOwnedOperations();
            else
                operations = ((Interface) inElt).getOwnedOperations();
            for (Object curOp_o : operations) {
                Operation curOp = (Operation) curOp_o;
                for (Object curParam_o : curOp.getOwnedParameters()) {
                    Parameter curParam = (Parameter) curParam_o;
                    Type paramType = curParam.getType();
                    if (paramType != null && belongsTo(paramType, inNS)) return true;
                }
            }
        }
        return false;
    }

    /**
     * Determines whether or not the input element belongs to the target package.
     */
    private static boolean belongsTo(NamedElement inElt, Namespace inTo) {
        Package container = inElt.getNearestPackage();
        if (container != null) {
            String containerName = getNameWithoutModel(container.getQualifiedName());
            String targetName = getNameWithoutModel(inTo.getQualifiedName());
            if (containerName.equals(targetName) || containedIn(containerName, targetName)) {
                return true;
            }
        }
        return false;
    }

    public static IResource getModelResource(NamedElement elt) {
        Model model;
        if (elt instanceof Model)
            model = (Model) elt;
        else
            model = elt.getModel();
        Resource r = model.eResource();
        if (r != null) {
            IWorkspaceRoot wr = ResourcesPlugin.getWorkspace().getRoot();
            String uri = r.getURI().path().replaceAll("%20", " ");
            if (uri.startsWith("/resource/")) {
                String pName = uri.substring("/resource".length() + 1);
                String mName = null;
                int pos = pName.indexOf("/");
                if (pos > 0) {
                    mName = pName.substring(pos + 1);
                    pName = pName.substring(0, pos);
                    IProject prj = wr.getProject(pName);
                    if (prj != null) {
                        return prj.findMember(mName);
                    }
                }
            } else {
                // The following code handles the situation when a PathMap entry is defined: in this
                // case the uri is something like //PATHMAP_ENTRY/fileName.emx. Because I don't know
                // how to get to the Path Map entry, I came up with this awkward code: go through
                // all projects in the workspace, then locate the files 'fileName.emx'. If there are
                // duplicates, retrieve the corresponding model and look if it is equal to the input
                // model.
                int pos = uri.lastIndexOf("/");
                String fName = uri;
                if (pos >= 0) fName = fName.substring(pos + 1);
                List<IResource> resources = new ArrayList<IResource>();
                for (int i = 0; i < wr.getProjects().length; i++) {
                    IProject prj = wr.getProjects()[i];
                    try {
                        for (int j = 0; j < prj.members().length; j++) {
                            IResource rsc = prj.members()[j];
                            if (rsc.getName().equals(fName)) resources.add(rsc);
                        }
                    } catch (CoreException e) {
                        System.err.println("RSMModel.getModelResource: Cannot retrieve the projects from the current workspace.");
                    }
                }
                if (!resources.isEmpty()) {
                    if (resources.size() == 1) {
                        return (IResource) resources.get(0);
                    } else {
                        for (Iterator i = resources.iterator(); i.hasNext();) {
                            IResource rsc = (IResource) i.next();
                            IFile f = (IFile) rsc;
                            Model m;
                            try {
                                m = UMLModeler.openModel(wr.getLocation().toOSString() + f.getFullPath().toOSString());
                                if (m.equals(model)) {
                                    return (IResource) rsc;
                                }
                            } catch (IOException e) {
                                System.err.println("RSMModel.getModelResource: unable to open model for file "
                                        + wr.getLocation().toOSString() + f.getFullPath().toOSString());
                            }
                        }
                    }
                }
            }
        }
        return null;
    }

    public static IMarker addMarker(IResource onRsrc, String description, int severity) {
        if (onRsrc == null) return null;
        try {
            IMarker marker = onRsrc.createMarker(IMarker.PROBLEM);
            if (marker.exists()) {
                marker.setAttribute(IMarker.MESSAGE, description);
                marker.setAttribute(IMarker.SEVERITY, severity);
                marker.setAttribute(IMarker.TRANSIENT, true);
                return marker;
            }
        } catch (CoreException e) {
            System.err.println("RSMModel.addMarker: unable to create marker.");
        }
        return null;
    }

    public static IProject getProject(String projectName) {
        IWorkspaceRoot wr = ResourcesPlugin.getWorkspace().getRoot();
        for (int i = 0; i < wr.getProjects().length; i++) {
            IProject prj = wr.getProjects()[i];
            try {
                for (int j = 0; j < prj.members().length; j++) {
                    IResource rsc = prj.members()[j];
                    if (rsc.getName().equals(projectName)) {
                        return prj;
                    }
                }
            } catch (CoreException e) {
                System.err.println("RSMModel.getProject: Cannot retrieve the projects from the current workspace.");
            }
        }
        return null;
    }

    public static IProject getProject(NamedElement elt) {
        Model model;
        if (elt instanceof Model)
            model = (Model) elt;
        else
            model = elt.getModel();
        Resource r = model.eResource();
        if (r != null) {
            IWorkspaceRoot wr = ResourcesPlugin.getWorkspace().getRoot();
            String uri = r.getURI().path().replaceAll("%20", " ");
            if (uri.startsWith("/resource/")) {
                String pName = uri.substring("/resource".length() + 1);
                int pos = pName.indexOf("/");
                if (pos > 0) {
                    pName = pName.substring(0, pos);
                    IProject prj = wr.getProject(pName);
                    return prj;
                }
            } else {
                // The following code handles the situation when a PathMap entry is defined: in this
                // case the uri is something like //PATHMAP_ENTRY/fileName.emx. Because I don't know
                // how to get to the Path Map entry, I came up with this awkward code: go through
                // all projects in the workspace, then locate the files 'fileName.emx'. If there are
                // duplicates, retrieve the corresponding model and look if it is equal to the input
                // model.
                int pos = uri.lastIndexOf("/");
                String fName = uri;
                if (pos >= 0) fName = fName.substring(pos + 1);
                List<IResource> resources = new ArrayList<IResource>();
                List<IProject> projects = new ArrayList<IProject>();
                for (int i = 0; i < wr.getProjects().length; i++) {
                    IProject prj = wr.getProjects()[i];
                    try {
                        for (int j = 0; j < prj.members().length; j++) {
                            IResource rsc = prj.members()[j];
                            if (rsc.getName().equals(fName)) {
                                projects.add(prj);
                                resources.add(rsc);
                            }
                        }
                    } catch (CoreException e) {
                        System.err
                                .println("RSMModel.getProject: Cannot retrieve the projects from the current workspace.");
                    }
                }
                if (!resources.isEmpty()) {
                    if (resources.size() == 1) {
                        return (IProject) projects.get(0);
                    } else {
                        for (int i = 0; i < resources.size(); i++) {
                            IResource rsc = (IResource) resources.get(i);
                            IFile f = (IFile) rsc;
                            Model m;
                            try {
                                m = UMLModeler.openModel(wr.getLocation().toOSString() + f.getFullPath().toOSString());
                                if (m.equals(model)) {
                                    return (IProject) projects.get(i);
                                }
                            } catch (IOException e) {
                                System.err.println("RSMModel.getProject: unable to open model for file "
                                        + wr.getLocation().toOSString() + f.getFullPath().toOSString());
                            }
                        }
                    }
                }
            }
        }
        return null;
    }

    public static Model getModel(Element elt) {
        if (elt instanceof Model)
            return (Model) elt;
        else
            return elt.getModel();
    }
    public static List<Element> findLinkedElement(Package inPkg, EClass inElementKind, EClass inRelationKind, String inElementName, boolean recursiveSearch) {
        List<Element> list = new ArrayList<Element>();
        if (recursiveSearch) {
        	findElements(list, inPkg, inElementKind, inElementName);
        } else {
        	Element elt = findElement(inPkg, inElementKind, inElementName);
        	if (elt != null) {
        		list.add(elt);
        	}
        }
        return list;
    }


    public static List<Element> findElements(Package inPkg, EClass inElementKind, String inElementName) {
        List<Element> list = new ArrayList<Element>();
        findElements(list, inPkg, inElementKind, inElementName);
        return list;
    }

    private static void findElements(List<Element> list, Package inPkg, EClass inElementKind, String inElementName) {
    	Element elt = findElement(inPkg, inElementKind, inElementName);
    	if (elt != null) {
    		list.add(elt);
    	}
    	for (Iterator i = inPkg.getNestedPackages().iterator(); i.hasNext();) {
            findElements(list, (Package) i.next(), inElementKind, inElementName);
        }
    }

    public static Element findElement(Package inPkg, EClass inElementKind, String inElementName) {
        for (Object curElt_o:inPkg.getOwnedElements()) {
            if (curElt_o instanceof NamedElement) {
                NamedElement curElt = (NamedElement) curElt_o;
                if (curElt.eClass() == inElementKind && curElt.getName().equals(inElementName)) {
                    return curElt;
                }
            }
        }
        return null;
    }  

	public static List<EObject> findElements(EObject inObj, EClass inElementKind) {
		List<EObject> temp = new ArrayList<EObject>();
		for (Object curObj_o:inObj.eContents()) {
			EObject curObj = (EObject) curObj_o;
			if (inElementKind.isSuperTypeOf(curObj.eClass())) {
				temp.add((EObject) curObj);
			}
			temp.addAll(findElements((EObject) curObj_o, inElementKind));
		}
		return temp;
	}
	

	@SuppressWarnings("unchecked")
	public static List<RSMRelation> findClients(EObject inObj, Classifier inCl) {
		List<RSMRelation> list = new ArrayList<RSMRelation>();
		List<EObject> temp = findElements(inObj, UMLPackage.eINSTANCE.getDirectedRelationship());
		for (EObject curElt:temp) {
			DirectedRelationship rel = (DirectedRelationship) curElt;
			for (Object curTgt_o:rel.getTargets()) {
				Element curTgt = (Element) curTgt_o;
				if (curTgt != null && curTgt == inCl && !rel.getSources().isEmpty()) {
					RSMRelation res = new RSMRelation();
					res.setClient((NamedElement) rel.getSources().get(0));
					res.setRel(rel);
					res.setSupplier(inCl);
					list.add(res);
					break;
				}
			}
		}

		temp = findElements(inObj, UMLPackage.eINSTANCE.getAssociation());
		for (EObject curElt:temp) {
			Association assoc = (Association) curElt;
			EList members = assoc.getMemberEnds();
			if (members.size() >= 2) {
				Property prop1 = (Property) members.get(0);
				Property prop2 = (Property) members.get(1);
				Type type1 = prop1.getType();
				Type type2 = prop2.getType();
				if (type1 instanceof Classifier && type2 instanceof Classifier) {
					Classifier cl1 = (Classifier) type1;
					Classifier cl2 = (Classifier) type2;
					Classifier client = null;
					Property prop = null;
					if (cl1.equals(inCl)) {
						client = cl2;
						prop = prop1;
					} else if (cl2.equals(inCl)) {
						client = cl1;
						prop = prop2;
					}
					if (prop != null && !assoc.getOwnedEnds().contains(prop)) {
						RSMRelation res = new RSMRelation();
						res.setClient(client);
						res.setRel(assoc);
						res.setSupplier(inCl);
						list.add(res);
					}
				}
			}
		}
		return list;
	}
    
	public static List<RSMRelation> findSuppliers(EObject inObj, Classifier inCl) {
		List<RSMRelation> list = new ArrayList<RSMRelation>();
		List<EObject> temp = findElements(inObj, UMLPackage.eINSTANCE.getDirectedRelationship());
		for (EObject curElt:temp) {
			DirectedRelationship rel = (DirectedRelationship) curElt;
			for (Object curSrc_o:rel.getSources()) {
				Element curSrc = (Element) curSrc_o;
				if (curSrc != null && curSrc == inCl && !rel.getTargets().isEmpty()) {
					RSMRelation res = new RSMRelation();
					res.setClient(inCl);
					res.setRel(rel);
					res.setSupplier((NamedElement) rel.getTargets().get(0));
					list.add(res);
					break;
				}
			}
		}

		temp = findElements(inObj, UMLPackage.eINSTANCE.getAssociation());
		for (EObject curElt:temp) {
			Association assoc = (Association) curElt;
			EList members = assoc.getMemberEnds();
			if (members.size() >= 2) {
				Property prop1 = (Property) members.get(0);
				Property prop2 = (Property) members.get(1);
				Type type1 = prop1.getType();
				Type type2 = prop2.getType();
				if (type1 instanceof Classifier && type2 instanceof Classifier) {
					Classifier cl1 = (Classifier) type1;
					Classifier cl2 = (Classifier) type2;
					Classifier target = null;
					Property prop = null;
					if (cl1.equals(inCl)) {
						target = cl2;
						prop = prop2;
					} else if (cl2.equals(inCl)) {
						target = cl1;
						prop = prop1;
					}
					if (prop != null && !assoc.getOwnedEnds().contains(prop)) {
						RSMRelation res = new RSMRelation();
						res.setClient(inCl);
						res.setRel(assoc);
						res.setSupplier(target);
						list.add(res);
					}
				}
			}
		}
		return list;
	}
    
	@SuppressWarnings("unchecked")
	public static List<View> findViews(EObject inContainer, Element inRef) {
		List<View> list = new ArrayList<View>();
		List<EObject> temp = findElements(inContainer, NotationPackage.eINSTANCE.getView());
		for (EObject curElt:temp) {
			View v = (View) curElt;
			EObject e = v.getElement();
			if (e != null && e == inRef) {
				list.add(v);
			}
		}
		return list;
	}
    
    public static void copyOperation(Operation op, Interface cl) {
        Operation targetOp = cl.createOwnedOperation(op.getName(), null, null);
        targetOp.setName(op.getName());
        targetOp.setIsAbstract(op.isAbstract());
        targetOp.setIsStatic(op.isStatic());
        targetOp.setType(op.getType());
        targetOp.setVisibility(op.getVisibility());
        setDocumentation(targetOp, getDocumentation(op));
        for (Object curParam_o:op.getOwnedParameters()) {
            Parameter curParam = (Parameter) curParam_o;
            Parameter targetParam = null;
            if (curParam.getDirection().getValue() != ParameterDirectionKind.RETURN) {
                targetParam = targetOp.createOwnedParameter(curParam.getName(), curParam.getType());
                targetParam.setDefaultValue(curParam.getDefaultValue());
                targetParam.setVisibility(curParam.getVisibility());
                targetParam.setLowerValue(curParam.getLowerValue());
                targetParam.setUpperValue(curParam.getUpperValue());
            } else {
                targetParam = targetOp.createReturnResult(curParam.getName(), curParam.getType());
            }
            targetParam.setDirection((curParam.getDirection()));
            setDocumentation(targetParam, getDocumentation(curParam));
        }
    }
    
    /**
     * Checks if a class contains the specified operation
     * 
     * @param cls
     *            The class to be searched for the presence of the specified operation.
     * @param op
     *            The operation
     * @return True if the class contains the operation.
     */
    public static boolean classContainsOperation(Class cls, Operation op) {
        boolean foundOperation = false;
        Iterator iter = cls.getOwnedOperations().iterator();
        while (foundOperation == false && iter.hasNext()) {
            Operation ownedOp = (Operation) iter.next();
            foundOperation = areOperationsEqual(ownedOp, op);
        }
        return foundOperation;
    }

    /**
     * Checks if a interface contains the specified operation
     * 
     * @param cls
     *            The interface to be searched for the presence of the specified operation.
     * @param inOp
     *            The operation
     * @return True if the interface contains the operation.
     */
    public static boolean interfaceContainsOperation(Interface inInt, Operation inOp) {
        boolean foundOperation = false;
        Iterator iter = inInt.getOwnedOperations().iterator();
        while (foundOperation == false && iter.hasNext()) {
            Operation ownedOp = (Operation) iter.next();
            foundOperation = areOperationsEqual(ownedOp, inOp);
        }
        return foundOperation;
    }

    /**
     * Checks if two operations are equal.
     * 
     * Two operations are defined to be equal if they have the same name and the same signature.
     * 
     * @param op1
     *            Operation 1
     * @param op2
     *            Operation 2
     * @return True if the operations have the same name and signature.
     */
    public static boolean areOperationsEqual(Operation op1, Operation op2) {
        return (op1.getName().equals(op2.getName())) && haveSameSignature(op1, op2);
    }

    /**
     * Checks if two operations have the same signature.
     * 
     * Two operations are defined to have the same signature if the parameter lists of the
     * operations have the same number of parameters, the parameters appear in the lists in the same
     * order, and each parameter in one operation has the same signature as the corresponding
     * parameter of the second operation. Note that the return parameter does not contribute to the
     * signature of an operation.
     * 
     * @param op1
     *            First operation
     * @param op2
     *            Second operation
     * @return True if the operations have the same signature
     */
    public static boolean haveSameSignature(Operation op1, Operation op2) {
        boolean isSame = false;
        // To ignore return parameters, create a list of non-return parameters
        // for each operation
        List op1Parameters = getNonReturnParameters(op1);
        List op2Parameters = getNonReturnParameters(op2);
        if (op1Parameters.size() == op2Parameters.size()) {
            Iterator op1paramIter = op1Parameters.iterator();
            Iterator op2paramIter = op2Parameters.iterator();
            isSame = true;
            while (op1paramIter.hasNext() && isSame == true) {
                isSame = haveSameSignature((Parameter) op1paramIter.next(), (Parameter) op2paramIter.next());
            }
        }
        return isSame;
    }

    /**
     * Checks if two parameters have the same signature.
     * 
     * Two parameters are defined to have the same signature if they are of the same type and they
     * have the same direction. The names of the parameters are ignored because in the context of an
     * operation the names do not determine the signature of an operation.
     * 
     * Also, if the types of the parameters are unspecified (that is if getType() for each parameter
     * is <code>null</code>, the parameters are defined to have the same signature only if their
     * directions are the same.
     * 
     * @param p1
     *            First parameter
     * @param p2
     *            Second parameter
     * @return True if the parameters have the same types and direction
     */
    public static boolean haveSameSignature(Parameter p1, Parameter p2) {
        boolean isSame = false;
        Type type1 = p1.getType();
        Type type2 = p2.getType();
        if (p1.getDirection() == p2.getDirection()) {
            if (type1 == null && type2 == null) {
                isSame = true;
            } else if (type1 != null && type2 != null) {
                isSame = type1.eClass() == type2.eClass();
            }
        }
        return isSame;
    }


    /**
     * Creates a list of non-return (in, out, and in-out) parameters of an operation.
     * 
     * @param op
     *            The specified operation
     * @return List of non-return parameters.
     */
    private static List getNonReturnParameters(Operation op) {
        List<Parameter> plist = new ArrayList<Parameter>();
        Iterator iter = op.getOwnedParameters().iterator();
        while (iter.hasNext()) {
            Parameter p = (Parameter) iter.next();
            if (p.getDirection() != ParameterDirectionKind.RETURN_LITERAL) {
                plist.add(p);
            }
        }
        return plist;
    }

	public static String getBasicInfo(EObject elt) {
		String s = "";
		if (elt != null) {
			s = elt.getClass().getSimpleName();
			if (s.endsWith("Impl"))
				s = s.substring(0, s.length() - "Impl".length());
			else
				s = elt.getClass().getSimpleName();
			if (elt instanceof NamedElement) {
				NamedElement ne = (NamedElement) elt;
				if (ne.getName() == null || ne.getName().equals(""))
					s += " <unnamed>";
				else
					s += " '" + ne.getName() + "'";
				if (ne instanceof Message) {
					Message msg = (Message) ne;
					s += " (" + msg.getMessageSort().getLiteral() + ")";
				}
			} else if (elt instanceof Node) {
				Node n = (Node) elt;
				EObject e = n.getElement();
				if (e != null) {
					s += " (" + getBasicInfo(e) + ")";
				}
			} else if (elt instanceof Edge) {
				Edge ed = (Edge) elt;
				EObject e = ed.getElement();
				if (e != null) {
					s += " (" + getBasicInfo(e) + ")";
				}				
			}
		} else {
			s = "element = <null>";
		}
		return s;
	}
	
	public static boolean isGlobal(Element inElt) {
		return (inElt.getAppliedStereotype("ArchitectAssistant::global") != null);
	}

	public static boolean isExternalLibrary(Package inPkg) {
		Package p = inPkg;
		Package parent = null;
		while (true) {
			if (p.getAppliedStereotype("ArchitectAssistant::externalLibrary") != null || 
				p.getAppliedStereotype("Standard::modelLibrary") != null) return true;
			parent = p.getNestingPackage();
			if (parent == null || parent instanceof Model) return false;
			p = parent;
		}
	}


}
